from .collector import Collector
from .exceptions import ManifestClientError
from .manifest import HashableResource, Manifests
from .manipulations import (
    Addition,
    ConfigRegistry,
    CreateNamespace,
    ManifestLabel,
    Patch,
    SubtractEq,
    update_tolerations,
)

__all__ = [
    "Addition",
    "Collector",
    "ConfigRegistry",
    "CreateNamespace",
    "HashableResource",
    "ManifestLabel",
    "Manifests",
    "Patch",
    "update_tolerations",
    "SubtractEq",
    "ManifestClientError",
]



# Copyright 2022 Canonical Ltd.
# See LICENSE file for licensing details.
"""Classes used for mutating or adding to manifests."""

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable, Iterable, List, Mapping, Optional

from lightkube import codecs
from lightkube.codecs import AnyResource
from lightkube.generic_resource import GenericGlobalResource, GenericNamespacedResource
from lightkube.models.core_v1 import Toleration
from lightkube.models.meta_v1 import Time

if TYPE_CHECKING:
    from .manifest import Manifests  # pragma: no cover

log = logging.getLogger(__file__)


@dataclass
class AnyCondition:
    """Condition describes the state of a resources at a certain point.

    **parameters**

    * **status** ``str`` - Status of the condition, one of True, False, Unknown.
    * **type** ``str`` - Type of replica set condition.
    * **lastTransitionTime** ``meta_v1.Time`` - *(optional)* The last time the condition
                                                transitioned from one status to another.
    * **message** ``str`` - *(optional)* A human readable message indicating details
                                         about the transition.
    * **reason** ``str`` - *(optional)* The reason for the condition's last transition.
    """

    status: "str"
    type: "str"
    lastTransitionTime: Optional[Time] = None
    message: Optional[str] = None
    reason: Optional[str] = None


def _unique(collection, key):
    """Yields a unique iterable of items from collection.

    uniqueness is determined from the result of key(item).
    """
    seen = set()

    for item in collection:
        value = key(item)
        if value not in seen:
            seen.add(value)
            yield item


class HashableResource:
    """Wraps a lightkube resource object so it is hashable."""

    def __init__(self, resource: AnyResource):
        self.resource = resource

    def __uniq(self):
        return self.kind, self.namespace, self.name

    @staticmethod
    def _condition_unwrap(condition: Mapping[str, str]) -> Optional[AnyCondition]:
        """Attempt to retrieve status and type from a Mapping"""
        try:
            _status, _type = (condition[_] for _ in ("status", "type"))
            return AnyCondition(_status, _type)
        except KeyError:
            return None

    @property
    def status_conditions(self) -> List[AnyCondition]:
        status = getattr(self.resource, "status", None)
        if not status:
            return []
        elif isinstance(self.resource.status, dict):
            conditions = [
                _
                for c in self.resource.status.get("conditions", [])
                for _ in map(self._condition_unwrap, [c])
                if _
            ]
        else:
            conditions = getattr(self.resource.status, "conditions", [])
        return conditions

    @property
    def kind(self) -> str:
        """Return the resource's kind."""
        return self.resource.kind or type(self.resource).__name__

    @property
    def namespace(self) -> Optional[str]:
        """Return the resource's namespace."""
        return self.resource.metadata.namespace if self.resource.metadata else None

    @property
    def name(self) -> Optional[str]:
        """Return the resource's name."""
        return self.resource.metadata.name if self.resource.metadata else None

    def __str__(self):
        """String version of the unique parts.

        example: 'kind/[namespace/]name'
        """
        return "/".join(filter(None, self.__uniq()))

    def __hash__(self):
        """Returns a hash of the unique parts."""
        return hash(self.__uniq())

    def __eq__(self, other):
        """Comparison only of the unique parts."""
        return isinstance(other, HashableResource) and other.__uniq() == self.__uniq()


class Manipulation:
    """Class used to support charm deviations from the manifests."""

    def __init__(self, manifests: "Manifests") -> None:
        self.manifests = manifests


class Patch(Manipulation):
    """Class used to define how to patch an existing object in the manifests."""

    def __call__(self, obj: AnyResource) -> None:
        """Method called to optionally update the object before application."""
        ...


class Addition(Manipulation):
    """Class used to define objects to add to the original manifests."""

    def __call__(self) -> Optional[AnyResource]:
        """Method called to optionally create an object."""
        ...


class Subtraction(Manipulation):
    """Class used to define objects to subtract from the original manifests."""

    def __call__(self, obj: AnyResource) -> bool:  # type: ignore
        """Method called to optionally subtract an object.

        this is an abstract method, each implementation should return a bool
        """
        ...


class CreateNamespace(Addition):
    """Class used to create additional namespace before apply manifests."""

    def __init__(self, manifests: "Manifests", namespace: str) -> None:
        super().__init__(manifests)
        self.namespace = namespace

    def __call__(self) -> Optional[AnyResource]:
        """Create the default namespace if available."""
        log.info(f"Creating namespace {self.namespace}")
        return codecs.from_dict(
            dict(
                apiVersion="v1",
                kind="Namespace",
                metadata=dict(name=self.namespace),
            )
        )


class ManifestLabel(Patch):
    """Ensure every manifest item is labeled with the manifest name.

    Similar to helm charts, add to each metadata some information
    regarding what applied this resource up
    https://helm.sh/docs/chart_best_practices/labels/
    """

    def __call__(self, obj: AnyResource):
        """Adds manifest.name label to obj."""
        if obj.metadata:
            version = self.manifests.current_release
            labels = {
                "juju.io/application": self.manifests.model.app.name,
                "juju.io/manifest": self.manifests.name,
                "juju.io/manifest-version": f"{self.manifests.name}-{version}",
            }
            if isinstance(obj, (GenericGlobalResource, GenericNamespacedResource)):
                # Custom resources in lightkube are built differently
                # from standard model resources
                obj["metadata"]["labels"] = obj.metadata.labels or {}
            else:
                # ensure object has labels
                obj.metadata.labels = obj.metadata.labels or {}
            obj.metadata.labels.update(**labels)


class ConfigRegistry(Patch):
    """Applies image registry to the manifest."""

    def __call__(self, obj):
        """Use the image-registry config and updates container images in obj."""
        registry = self.manifests.config.get("image-registry")
        if not registry:
            return
        if obj.kind in [
            # https://kubernetes.io/docs/concepts/workloads/pods/
            "Pod"
        ]:
            spec = obj.spec
        elif obj.kind in [
            # https://kubernetes.io/docs/concepts/workloads/controllers/daemonset/
            "DaemonSet",
            # https://kubernetes.io/docs/concepts/workloads/controllers/deployment/
            "Deployment",
            # https://kubernetes.io/docs/concepts/workloads/controllers/job/
            "Job",
            # https://kubernetes.io/docs/concepts/workloads/controllers/replicaset/
            "ReplicaSet",
            # https://kubernetes.io/docs/concepts/workloads/controllers/replicationcontroller/
            "ReplicationController",
            # https://kubernetes.io/docs/concepts/workloads/controllers/statefulset/
            "StatefulSet",
        ]:
            spec = obj.spec.template.spec
        elif obj.kind in [
            # https://kubernetes.io/docs/concepts/workloads/controllers/cron-jobs/
            "CronJob"
        ]:
            spec = obj.spec.jobTemplate.spec.template.spec
        else:
            spec = None

        containers = []
        if spec:
            if spec.containers:
                containers += spec.containers
            if spec.initContainers:
                containers += spec.initContainers

        for container in containers:
            full_image = container.image
            if full_image:
                _, image = full_image.split("/", 1)
                new_full_image = f"{registry}/{image}"
                container.image = new_full_image
                log.info(f"Replacing Image: {full_image} with {new_full_image}")


TolerationAdjuster = Callable[[List[Toleration]], Iterable[Toleration]]


def update_tolerations(obj: AnyResource, adjuster: TolerationAdjuster):
    """Uses the adjuster service and updates any object tolerations."""
    if obj.kind in ["Pod"]:
        spec = obj.spec
    elif obj.kind in ["DaemonSet", "Deployment", "StatefulSet"]:
        spec = obj.spec.template.spec
    else:
        spec = None

    if spec:
        updated = list(
            _unique(
                adjuster(spec.tolerations), key=lambda t: tuple(t.to_dict().values())
            )
        )
        log.info(f"Applying tolerations {updated} to {HashableResource(obj)}")
        spec.tolerations = updated
    return obj


class SubtractEq(Subtraction):
    """Ensure every manifest item is labeled with the manifest name.

    Similar to helm charts, add to each metadata some information
    regarding what applied this resource up
    https://helm.sh/docs/chart_best_practices/labels/
    """

    def __init__(self, manifests: "Manifests", to_compare: AnyResource) -> None:
        super().__init__(manifests)
        self.to_compare = to_compare

    def __call__(self, obj: AnyResource) -> bool:
        """Returns true if obj == rsc based on kind, name, and namespace"""
        return HashableResource(self.to_compare) == HashableResource(obj)



# Copyright 2022 Canonical Ltd.
# See LICENSE file for licensing details.

from collections import OrderedDict
from dataclasses import dataclass
from typing import Iterable, List, Mapping, MutableMapping, Optional, Tuple

from .manifest import Manifests
from .manipulations import AnyCondition, HashableResource


@dataclass
class _ResourceAnalysis:
    correct: Iterable[HashableResource] = frozenset()
    extra: Iterable[HashableResource] = frozenset()
    missing: Iterable[HashableResource] = frozenset()


class Collector:
    def __init__(self, *manifests: Manifests) -> None:
        """Collect manifests for an ops charm.

        Usage:
        from ops.manifests import Collector

        class ExampleCharm(CharmBase):
            def __init__(self, *args):
                super().__init__(*args)

                # collection of ManifestImpls
                self.manifests = [ ... ]

                # create an Collector handler object
                self.collector = Collector(*self.manifests)

                # Register collector callbacks
                self.framework.observe(
                    self.on.list_versions_action,
                    self.collector.list_versions
                )
        """
        d = {manifest.name: manifest for manifest in manifests}
        self.manifests = OrderedDict(sorted(d.items(), key=lambda x: x[0]))

    def list_versions(self, event) -> None:
        """Respond to list_versions action."""
        result = {
            f"{name}-versions": "\n".join(str(_) for _ in manifest.releases)
            for name, manifest in self.manifests.items()
        }
        event.set_results(result)

    def list_resources(self, event, manifests: Optional[str], resources: Optional[str]):
        """List available, extra, and missing resources for each manifest."""
        self._list_resources(event, manifests, resources)

    def scrub_resources(
        self, event, manifests: Optional[str], resources: Optional[str]
    ):
        """Remove extra resources installed by each manifest.

        Uses the list_resource analysis to determine the extra resource
        then delete those resources.
        """
        results = self._list_resources(event, manifests, resources)
        for name, analysis in results.items():
            if analysis.extra:
                event.log(f"Removing {','.join(str(_) for _ in analysis.extra)}")
                self.manifests[name].delete_resources(*analysis.extra)
        self._list_resources(event, manifests, resources)

    def apply_missing_resources(
        self, event, manifests: Optional[str], resources: Optional[str]
    ):
        """Applies manifest resources that are missing from the cluster

        Uses the list_resource analysis to determine the missing resources
        then applies those resources.
        """
        results = self._list_resources(event, manifests, resources)
        for name, analysis in results.items():
            if analysis.missing:
                event.log(f"Applying {','.join(str(_) for _ in analysis.missing)}")
                self.manifests[name].apply_resources(*analysis.missing)
        self._list_resources(event, manifests, resources)

    @property
    def unready(self) -> List[str]:
        """List of statuses of resources with non-ready conditions."""
        return sorted(
            f"{name}: {obj} is not {cond.type}"
            for (name, obj), cond in self.conditions.items()
            if cond.status != "True"
        )

    @property
    def conditions(self) -> Mapping[Tuple[str, HashableResource], AnyCondition]:
        """
        Condition of all resources with `$obj.status.conditions[]`.

        The key of the mapping is the pair of
          * the name of the manifest in the collection
          * the installed resource in the cluster

        Each value represents a Condition which conforms to proposed spec
        https://github.com/kubernetes/enhancements/blob/dfb5b64322fe861fcc173ca8246a2ec5a511e46e/keps/sig-api-machinery/1623-standardize-conditions/README.md#proposal
        """
        return {
            (name, obj): cond
            for name, manifest in self.manifests.items()
            for obj in manifest.status()
            for cond in obj.status_conditions
        }

    @property
    def short_version(self) -> str:
        """Short status of collective manifests."""
        return ",".join(c.current_release for c in self.manifests.values())

    @property
    def long_version(self) -> str:
        """Long status of collective manifests."""
        return "Versions: " + ", ".join(
            f"{app}={c.current_release}" for app, c in self.manifests.items()
        )

    def _list_resources(
        self, event, manifests: Optional[str], resources: Optional[str]
    ):
        filter_manifests = manifests.split() if manifests else []
        filter_resources = resources.split() if resources else []

        man_filter = set(_.lower() for _ in filter_manifests)
        if man_filter:
            event.log(f"Filter manifest listings with {man_filter}")
        man_filter = man_filter or set(self.manifests.keys())

        res_filter = set(_.lower() for _ in filter_resources)
        if res_filter:
            event.log(f"Filter resource listing with {res_filter}")

        def kind_filter(rsc: HashableResource) -> bool:
            return not res_filter or rsc.kind.lower() in res_filter

        results: MutableMapping[str, _ResourceAnalysis] = {}
        event_result: MutableMapping[str, str] = {}
        for name, manifest in self.manifests.items():
            if name not in man_filter:
                results[name] = _ResourceAnalysis()
                continue

            labelled = manifest.labelled_resources()
            expected = manifest.resources
            installed = manifest.installed_resources()

            analyses = [expected & installed, labelled - expected, expected - installed]
            analyses = [frozenset(filter(kind_filter, cws)) for cws in analyses]
            correct, extra, missing = analyses

            results[name] = _ResourceAnalysis(correct, extra, missing)
            event_result.update(
                {
                    f"{name}-correct": "\n".join(sorted(str(_) for _ in correct)),
                    f"{name}-extra": "\n".join(sorted(str(_) for _ in extra)),
                    f"{name}-missing": "\n".join(sorted(str(_) for _ in missing)),
                }
            )

        event_result = {k: v for k, v in event_result.items() if v}
        event.set_results(event_result)
        return results
